"""
Text DNN is the DNN based on predefined embedding word vectors.
"""
from gensim.models import KeyedVectors
from pathlib import Path
import pandas as pd
import spacy
from ..utils import nlp
import numpy as np
from keras import utils
from time import time


DEFAULT_LANGUAGE_MODEL = "en_core_web_sm"


class TextDNNDataGenerator(utils.Sequence):
    """
    Read data from given file, and generate training data
    File format:
    - csv format, separated by comma
    - 1st column is the text
    - 2nd column is class labels, if multiple label, using space as separator
    """
 
    def __init__(self, datafile:Path, embeddings, batch_size, n_classes, random_seed=None, shuffle=True, language_model=DEFAULT_LANGUAGE_MODEL):
    
        self.shuffle = shuffle
        self.batch_size = batch_size
        self.n_classes = n_classes
        self.random_seed = random_seed or int(time())

        np.random.seed(self.random_seed)
        self._prepare_embeddings(embeddings)
        self._prepare_data(datafile, language_model)

        self._prepare_nlps(datafile, language_model)
        self._prepare_trainset()
        self.on_epoch_end()

    def _prepare_embeddings(self, embeddings):
        if isinstance(embeddings, KeyedVectors):
            self.embeddings = embeddings
        elif isinstance(embeddings, Path):
            self.embeddings = KeyedVectors.load_word2vec_format(embeddings.as_posix())
        else:
            raise AssertionError(f'embeddings should be KeyedVectors or Path')

    def _prepare_data(self, datafile, language_model):
        df = pd.read_csv(datafile)

        classes = set(map(str.strip, ' '.join(df.iloc[:, 1].tolist()).split(' ')))
        assert len(classes) <= self.n_classes, f'given n_classess={self.n_classes} while NO. of classes={len(classes)}.'
        self.classes = list(classes)
        
        doc2vec = lambda x: np.mean([self.embeddings[t.lemma_] for t in nlp.tokenize(x)], axis=0)
        to_index = lambda x: [self.classes.index(c.strip()) for c in x.split(' ')]
        onehot = lambda x: np.bitwise_or.reduce(utils.to_categorical(to_index(x), num_classes=self.n_classes, dtype='int32'), axis=0)

        df['vectors'] = df.iloc[:, 0].map(doc2vec)
        df['labels'] = df.iloc[:, 1].map(onehot)

        self.data = df[['vectors', 'labels']]
    
    def __len__(self):
        return self.data.shape[0]

    def __getitem__(self, index):

        istart, iend = index * self.batch_size, (index + 1) * self.batch_size
        
        return self.data.iloc[istart:iend, :]

        '''
        X = np.empty((len(batch_instance_ids), self.n_vocab_size))
        Y = np.empty((len(batch_instance_ids), self.n_vocab_size), dtype=int)
        for i, instance_id in enumerate(batch_instance_ids):
            X[i, ] = self._oh_encode_tokens([self.instances[i][0]])
            Y[i, ] = self._oh_encode_tokens(self.instances[i][1])

        return X, Y
        '''

    def on_epoch_end(self):
        if self.shuffle:
            self.data = self.data.sample(1)
        return super().on_epoch_end()

    def summary(self):
        return f'''>>> Skip Gram Data Generator Summary:
        | Name             |                       Value                        |
        |------------------+----------------------------------------------------|
        | n_vocab_size     | {self.n_vocab_size:^50d} |
        | batch_size       | {self.batch_size:^50d} |
        | r_sample         | {self.r_sample:^50f} |
        | n_window         | {self.n_window:^50d} |
        | steps_each_epoch | {self.n_vocab_size // self.batch_size + 1:^50d} |
        \n'''



from keras.layers import Dense, Input
from keras.models import Model
from pathlib import Path
import spacy
from itertools import product, chain
from keras.utils import Sequence
from collections import Counter
import numpy as np
from time import time
from keras import backend as K


class SkipGramDataGenerator(Sequence):
 
    def __init__(self, datafile:Path, batch_size, r_sample=0.001, n_window=2, language_model="xx_ent_wiki_sm", random_seed=None, shuffle=True):
    
        self.shuffle = shuffle
        self.n_window = n_window
        self.r_sample = r_sample
        self.batch_size = batch_size

        np.random.seed(random_seed or int(time()))
        self._prepare_nlps(datafile, language_model)
        self._prepare_trainset()
        self.on_epoch_end()

    def _prepare_nlps(self, datafile, language_model):
        nlp = spacy.load(language_model)
        nlp.add_pipe(nlp.create_pipe('sentencizer'))

        with open(datafile, 'r') as f:
            doc = nlp(f.read())

        tokens_cnter = Counter([t.text for t in doc])
        self.kv_token_freq = {t: v / sum(tokens_cnter.values()) for t, v in tokens_cnter.items()}
        self.tokens = list(tokens_cnter.keys())
        self.n_vocab_size = len(self.tokens)

        P_subsample = lambda token: (np.sqrt(self.kv_token_freq[token] / self.r_sample) + 1) * self.r_sample / self.kv_token_freq[token]
        kv_token__P_subsample = {t: P_subsample(t) for t in self.tokens}

        subsample_sentence = lambda sent: [t for t in sent if kv_token__P_subsample[t.text] >= np.random.rand()]
        self.doc = [subsample_sentence(sent) for sent in doc.sents]

    def _prepare_trainset(self):
    
        offsets = list(range(-1 * self.n_window, 0)) + list(range(1, self.n_window + 1))
        pair_locs = lambda loc_ceil: [(i, offset) for i, offset in product(range(loc_ceil), offsets) if i + offset >=0 and i + offset < loc_ceil]

        instances = dict()
        for sentence in self.doc:
            for i, offset in pair_locs(len(sentence)):
                centra = sentence[i].text
                surround = sentence[i + offset].text
                instances[centra] = instances.get(centra, []) + [surround]
        self.instances = list(instances.items())

        self.kv_surround_token__instance_id = dict()
        for i, (_, surrounds) in enumerate(self.instances):
            for s in surrounds:
                self.kv_surround_token__instance_id[s] = self.kv_surround_token__instance_id.get(s, []) + [i]
    
    def __len__(self):
        return len(self.instances)

    def __getitem__(self, index):

        pos_instance_ids = self._pick_instances(self.batch_size // 2)
        neg_instance_ids = self._negative_sampling(pos_instance_ids)
        batch_instance_ids = pos_instance_ids + neg_instance_ids
        if len(batch_instance_ids) < self.batch_size:
            batch_instance_ids += self._pick_instances(self.batch_size - len(batch_instance_ids))

        X = np.empty((len(batch_instance_ids), self.n_vocab_size))
        Y = np.empty((len(batch_instance_ids), self.n_vocab_size), dtype=int)
        for i, instance_id in enumerate(batch_instance_ids):
            X[i, ] = self._oh_encode_tokens([self.instances[i][0]])
            Y[i, ] = self._oh_encode_tokens(self.instances[i][1])

        return X, Y
    
    def _oh_encode_tokens(self, tokens):
        encoding = np.zeros(self.n_vocab_size)
        indexs = [self.tokens.index(t) for t in tokens]
        encoding[indexs] = 1
        return encoding

    def _pick_instances(self, size):
        unfetched_instance_ids = np.squeeze(self.__unfetched_instance_flags.nonzero())
        if self.shuffle:
            np.random.shuffle(unfetched_instance_ids)

        picked = unfetched_instance_ids[:size]
        self.__unfetched_instance_flags[picked] = 0
        return picked.tolist()

    def _negative_sampling(self, positive_instance_ids):
    
        pos_surrounds = chain(*[self.instances[i][1] for i in positive_instance_ids]) 

        neg_candidate_instance_ids = []
        for token in set(self.kv_surround_token__instance_id.keys()) - set(pos_surrounds):
            for i in self.kv_surround_token__instance_id[token]:
                if self.__unfetched_instance_flags[i]:
                    neg_candidate_instance_ids.append(i)
            
        if self.shuffle:
            np.random.shuffle(neg_candidate_instance_ids)

        candidates_ids = neg_candidate_instance_ids[:self.batch_size // 2]
        self.__unfetched_instance_flags[candidates_ids] = 0
        return candidates_ids

    def on_epoch_end(self):
        self.__unfetched_instance_flags = np.ones(len(self.instances))
        return super().on_epoch_end()

    def summary(self):
        return f'''>>> Skip Gram Data Generator Summary:
        | Name             |                       Value                        |
        |------------------+----------------------------------------------------|
        | n_vocab_size     | {self.n_vocab_size:^50d} |
        | batch_size       | {self.batch_size:^50d} |
        | r_sample         | {self.r_sample:^50f} |
        | n_window         | {self.n_window:^50d} |
        | steps_each_epoch | {self.n_vocab_size // self.batch_size + 1:^50d} |
        \n'''


class SkipGram(Model):

    def __init__(self, n_vocab_size, n_embedding):
        super(SkipGram, self).__init__(name='SkipGram')
        
        self.n_vocab_size = n_vocab_size
        self.n_embedding = n_embedding
        self._prepare_network()

    def _prepare_network(self):
        self.hidden_layer = Dense(self.n_embedding, name="hidden_layer")
        self.output_layer = Dense(self.n_vocab_size, name="output_layer")
        
    def call(self, inputs):
        return self.output_layer(self.hidden_layer(inputs))

    def get_embeddings(self, inputs):
        return  K.function([self.hidden_layer.input],[self.hidden_layer.output])([inputs])[0]



import spacy
from itertools import filterfalse


DEFAULT_LANGUAGE_MODEL = 'en_core_web_sm'


def tokenize(text, language_model=DEFAULT_LANGUAGE_MODEL, lemmatize=True, remove_stopwords=True):
    nlp = spacy.load(language_model)
    
    is_stopword = lambda x: remove_stopwords and x.norm_ in spacy.lang.en.stop_words.STOP_WORDS
    is_useless = lambda x: x.is_punct or x.is_digit or x.is_space or x.like_num or x.like_url or x.like_email
    is_removable = lambda x: is_stopword(x) or is_useless(x)

    token_value = lambda x: x.lemma_ if lemmatize else x.text
    yield from map(token_value, filterfalse(is_removable, nlp(text)))


def sentencizer(text, language_model=DEFAULT_LANGUAGE_MODEL):
    nlp = spacy.load(language_model)
    nlp.add_pipe(nlp.create_pipe('sentencizer'), before="parser")

    yield from map(lambda x: x.text.strip(), nlp(text).sents)



from . import LinearModel, DNN
from keras.layers import Input, Dense, Dropout, Concatenate
from keras.models import Model
import logging


logger = logging.getLogger(__name__)


def build(input_shape, output_shape, dtype, dense_layers=[1024, 512, 128], r_drop=0.5, name="wide_deep"):
    assert len(dense_layers) >= 2, 'invalid dense layers'

    x = Input(shape=input_shape, dtype=dtype, name="deep_wide_input")

    deep_model = DNN(input_shape, dense_layers[-1], dtype, output_activation=None, dense_layers=dense_layers[:-1], r_drop=r_drop, name="deep")
    deep_model.compile(loss='mae', optimizer='adam')
    yd = deep_model(x)

    xw = Concatenate()([x, yd])
    wide_model = LinearModel((input_shape[-1] + dense_layers[-1], ), output_shape, dtype, name="wide")
    wide_model.compile(loss='mae', optimizer='adam')
    y = wide_model(xw)
    
    return Model(x, y)

    





from keras.layers import Input, Dense
from keras import Model
import logging


logger = logging.getLogger(__name__)


def build(input_shape, output_shape, dtype, activation=None, name='linear_model'):
    x = Input(shape=input_shape, dtype=dtype, name="linear_input")
    y = Dense(output_shape, activation=activation, name="linear_dense")(x)
    return Model(input=x, output=y, name=name)





from . import LinearModel
from keras.layers import Input, Dense, Dropout
from keras.models import Model, Sequential
from itertools import chain
import logging


logger = logging.getLogger(__name__)


def build(input_shape, output_shape, dtype, output_activation="softmax", dense_layers=[1024, 512], r_drop=0.5, name="DNN"):
    assert len(dense_layers) >= 1, 'dense layers should not be empty!'

    x = Input(shape=input_shape, dtype=dtype, name="deep_input")

    h = x
    for i, v in enumerate(dense_layers):
        h = Dense(dense_layers[i], activation='relu', name=f"deep_dense_{i}")(h)
        h = Dropout(r_drop, name=f"deep_dropout_{i}")(h)

    y = Dense(output_shape, activation=output_activation, name="deep_output")(h)
    return Model(x, y, name=name)
    






import logging
from logging.config import dictConfig


dictConfig({
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "basic": {
            "format"  : "[%(asctime)s] [%(process)d:%(thread)d] [%(levelname)s] [%(name)s] %(filename)s:%(funcName)s:%(lineno)d %(message)s",
            "datefmt" : "%Y-%m-%d %H:%M:%S"
        }
    },
    "handlers": {
        "console": {
            "level": "WARN",
            "class": "logging.StreamHandler",
            "formatter": "basic",
            "stream": "ext://sys.stdout"
        },
    },
    "loggers": {
        "kerasmodels" : {
            "handlers" : ["console"],
            "propagate": "true",
            "level"    : "WARN"
        }
    }
})

logger = logging.getLogger(__name__)
logger.debug('kerasmodels logs will be printed in console')



import codecs
from setuptools import setup, find_packages


def read_file(filename, cb):
    with codecs.open(filename, 'r', 'utf8') as f:
        return cb(f)


setup(
    name='keras-models',
    version='0.0.5',

    packages=find_packages(),
    url='https://github.com/Marcnuth/Keras-Models',

    author='Marcnuth',
    author_email='hxianxian@gmail.com',

    license="Apache License 2.0",

    description="Keras Models Hub",
    long_description=read_file('README.md', lambda f: f.read()),
    long_description_content_type='text/markdown',
    install_requires=read_file('requirements.txt', lambda f: list(filter(bool, map(str.strip, f)))),

    classifiers=[
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries',
        'Programming Language :: Python :: 3'
    ]
)


# -*- coding: utf-8 -*-
'''VGG16-places365 model for Keras
This codes is forked from: https://github.com/GKalliatakis/Keras-VGG16-places365/blob/master/vgg16_places_365.py
To be re-wrote in Keras way in future.

# Reference:
- [Places: A 10 million Image Database for Scene Recognition](http://places2.csail.mit.edu/PAMI_places.pdf)
'''

import os

import warnings
import numpy as np
from PIL import Image
from cv2 import resize

from keras import backend as K
from keras.layers import Input
from keras.layers.core import Activation, Dense, Flatten
from keras.layers.pooling import MaxPooling2D
from keras.models import Model
from keras.layers import Conv2D
from keras.regularizers import l2
from keras.layers.core import Dropout
from keras.layers import GlobalAveragePooling2D
from keras.layers import GlobalMaxPooling2D
from keras_applications.imagenet_utils import _obtain_input_shape
from keras.engine.topology import get_source_inputs
from keras.utils.data_utils import get_file
from keras.utils import layer_utils
from keras.preprocessing import image
from keras.applications.imagenet_utils import preprocess_input

WEIGHTS_PATH = 'https://github.com/GKalliatakis/Keras-VGG16-places365/releases/download/v1.0/vgg16-places365_weights_tf_dim_ordering_tf_kernels.h5'
CLASS_LABELS = [
    'airfield', 'airplane_cabin', 'airport_terminal', 'alcove', 'alley', 'amphitheater', 'amusement_arcade', 'amusement_park', 'apartment_building/outdoor', 
    'aquarium', 'aqueduct', 'arcade', 'arch', 'archaelogical_excavation', 'archive', 'arena/hockey', 'arena/performance', 'arena/rodeo', 'army_base', 
    'art_gallery', 'art_school', 'art_studio', 'artists_loft', 'assembly_line', 'athletic_field/outdoor', 'atrium/public', 'attic', 'auditorium', 
    'auto_factory', 'auto_showroom', 'badlands', 'bakery/shop', 'balcony/exterior', 'balcony/interior', 'ball_pit', 'ballroom', 'bamboo_forest', 'bank_vault', 
    'banquet_hall', 'bar', 'barn', 'barndoor', 'baseball_field', 'basement', 'basketball_court/indoor', 'bathroom', 'bazaar/indoor', 'bazaar/outdoor', 'beach', 
    'beach_house', 'beauty_salon', 'bedchamber', 'bedroom', 'beer_garden', 'beer_hall', 'berth', 'biology_laboratory', 'boardwalk', 'boat_deck', 'boathouse', 
    'bookstore', 'booth/indoor', 'botanical_garden', 'bow_window/indoor', 'bowling_alley', 'boxing_ring', 'bridge', 'building_facade', 'bullring', 
    'burial_chamber', 'bus_interior', 'bus_station/indoor', 'butchers_shop', 'butte', 'cabin/outdoor', 'cafeteria', 'campsite', 'campus', 'canal/natural', 
    'canal/urban', 'candy_store', 'canyon', 'car_interior', 'carrousel', 'castle', 'catacomb', 'cemetery', 'chalet', 'chemistry_lab', 'childs_room', 
    'church/indoor', 'church/outdoor', 'classroom', 'clean_room', 'cliff', 'closet', 'clothing_store', 'coast', 'cockpit', 'coffee_shop', 'computer_room', 
    'conference_center', 'conference_room', 'construction_site', 'corn_field', 'corral', 'corridor', 'cottage', 'courthouse', 'courtyard', 'creek', 'crevasse', 
    'crosswalk', 'dam', 'delicatessen', 'department_store', 'desert/sand', 'desert/vegetation', 'desert_road', 'diner/outdoor', 'dining_hall', 'dining_room', 
    'discotheque', 'doorway/outdoor', 'dorm_room', 'downtown', 'dressing_room', 'driveway', 'drugstore', 'elevator/door', 'elevator_lobby', 'elevator_shaft', 
    'embassy', 'engine_room', 'entrance_hall', 'escalator/indoor', 'excavation', 'fabric_store', 'farm', 'fastfood_restaurant', 'field/cultivated', 'field/wild', 
    'field_road', 'fire_escape', 'fire_station', 'fishpond', 'flea_market/indoor', 'florist_shop/indoor', 'food_court', 'football_field', 'forest/broadleaf', 
    'forest_path', 'forest_road', 'formal_garden', 'fountain', 'galley', 'garage/indoor', 'garage/outdoor', 'gas_station', 'gazebo/exterior', 
    'general_store/indoor', 'general_store/outdoor', 'gift_shop', 'glacier', 'golf_course', 'greenhouse/indoor', 'greenhouse/outdoor', 'grotto', 
    'gymnasium/indoor', 'hangar/indoor', 'hangar/outdoor', 'harbor', 'hardware_store', 'hayfield', 'heliport', 'highway', 'home_office', 'home_theater', 
    'hospital', 'hospital_room', 'hot_spring', 'hotel/outdoor', 'hotel_room', 'house', 'hunting_lodge/outdoor', 'ice_cream_parlor', 'ice_floe', 'ice_shelf', 
    'ice_skating_rink/indoor', 'ice_skating_rink/outdoor', 'iceberg', 'igloo', 'industrial_area', 'inn/outdoor', 'islet', 'jacuzzi/indoor', 'jail_cell', 
    'japanese_garden', 'jewelry_shop', 'junkyard', 'kasbah', 'kennel/outdoor', 'kindergarden_classroom', 'kitchen', 'lagoon', 'lake/natural', 'landfill', 
    'landing_deck', 'laundromat', 'lawn', 'lecture_room', 'legislative_chamber', 'library/indoor', 'library/outdoor', 'lighthouse', 'living_room', 'loading_dock', 
    'lobby', 'lock_chamber', 'locker_room', 'mansion', 'manufactured_home', 'market/indoor', 'market/outdoor', 'marsh', 'martial_arts_gym', 'mausoleum', 'medina', 
    'mezzanine', 'moat/water', 'mosque/outdoor', 'motel', 'mountain', 'mountain_path', 'mountain_snowy', 'movie_theater/indoor', 'museum/indoor', 'museum/outdoor', 
    'music_studio', 'natural_history_museum', 'nursery', 'nursing_home', 'oast_house', 'ocean', 'office', 'office_building', 'office_cubicles', 'oilrig', 
    'operating_room', 'orchard', 'orchestra_pit', 'pagoda', 'palace', 'pantry', 'park', 'parking_garage/indoor', 'parking_garage/outdoor', 'parking_lot', 
    'pasture', 'patio', 'pavilion', 'pet_shop', 'pharmacy', 'phone_booth', 'physics_laboratory', 'picnic_area', 'pier', 'pizzeria', 'playground', 'playroom', 
    'plaza', 'pond', 'porch', 'promenade', 'pub/indoor', 'racecourse', 'raceway', 'raft', 'railroad_track', 'rainforest', 'reception', 'recreation_room', 
    'repair_shop', 'residential_neighborhood', 'restaurant', 'restaurant_kitchen', 'restaurant_patio', 'rice_paddy', 'river', 'rock_arch', 'roof_garden', 
    'rope_bridge', 'ruin', 'runway', 'sandbox', 'sauna', 'schoolhouse', 'science_museum', 'server_room', 'shed', 'shoe_shop', 'shopfront', 'shopping_mall/indoor', 
    'shower', 'ski_resort', 'ski_slope', 'sky', 'skyscraper', 'slum', 'snowfield', 'soccer_field', 'stable', 'stadium/baseball', 'stadium/football', 
    'stadium/soccer', 'stage/indoor', 'stage/outdoor', 'staircase', 'storage_room', 'street', 'subway_station/platform', 'supermarket', 'sushi_bar', 'swamp', 
    'swimming_hole', 'swimming_pool/indoor', 'swimming_pool/outdoor', 'synagogue/outdoor', 'television_room', 'television_studio', 'temple/asia', 'throne_room', 
    'ticket_booth', 'topiary_garden', 'tower', 'toyshop', 'train_interior', 'train_station/platform', 'tree_farm', 'tree_house', 'trench', 'tundra', 
    'underwater/ocean_deep', 'utility_room', 'valley', 'vegetable_garden', 'veterinarians_office', 'viaduct', 'village', 'vineyard', 'volcano', 
    'volleyball_court/outdoor', 'waiting_room', 'water_park', 'water_tower', 'waterfall', 'watering_hole', 'wave', 'wet_bar', 'wheat_field', 'wind_farm', 
    'windmill', 'yard', 'youth_hostel', 'zen_garden']


def VGG16_Places365():
    """Instantiates the VGG16-places365 architecture.

    Optionally loads weights pre-trained on Places. Note that when using TensorFlow,
    for best performance you should set `image_data_format="channels_last"` in your Keras config at ~/.keras/keras.json.

    The model and the weights are compatible with both TensorFlow and Theano. 
    The data format convention used by the model is the one specified in your Keras config file.

    # Returns
        A Keras model instance.

        """

    # Determine proper input shape
    input_shape = _obtain_input_shape(None, default_size=224,  min_size=48, data_format=K.image_data_format(), require_flatten=True)
    img_input = Input(shape=input_shape)


    # Block 1
    x = Conv2D(filters=64, kernel_size=3, strides=(1, 1), padding='same', kernel_regularizer=l2(0.0002), activation='relu', name='block1_conv1')(img_input)
    x = Conv2D(filters=64, kernel_size=3, strides=(1, 1), padding='same', kernel_regularizer=l2(0.0002), activation='relu', name='block1_conv2')(x)
    x = MaxPooling2D(pool_size=(2, 2), strides=(2, 2), name="block1_pool", padding='valid')(x)

    # Block 2
    x = Conv2D(filters=128, kernel_size=3, strides=(1, 1), padding='same', kernel_regularizer=l2(0.0002), activation='relu', name='block2_conv1')(x)
    x = Conv2D(filters=128, kernel_size=3, strides=(1, 1), padding='same', kernel_regularizer=l2(0.0002), activation='relu', name='block2_conv2')(x)
    x = MaxPooling2D(pool_size=(2, 2), strides=(2, 2), name="block2_pool", padding='valid')(x)

    # Block 3
    x = Conv2D(filters=256, kernel_size=3, strides=(1, 1), padding='same', kernel_regularizer=l2(0.0002), activation='relu', name='block3_conv1')(x)
    x = Conv2D(filters=256, kernel_size=3, strides=(1, 1), padding='same', kernel_regularizer=l2(0.0002), activation='relu', name='block3_conv2')(x)
    x = Conv2D(filters=256, kernel_size=3, strides=(1, 1), padding='same', kernel_regularizer=l2(0.0002), activation='relu', name='block3_conv3')(x)
    x = MaxPooling2D(pool_size=(2, 2), strides=(2, 2), name="block3_pool", padding='valid')(x)

    # Block 4
    x = Conv2D(filters=512, kernel_size=3, strides=(1, 1), padding='same', kernel_regularizer=l2(0.0002), activation='relu', name='block4_conv1')(x)
    x = Conv2D(filters=512, kernel_size=3, strides=(1, 1), padding='same', kernel_regularizer=l2(0.0002), activation='relu', name='block4_conv2')(x)
    x = Conv2D(filters=512, kernel_size=3, strides=(1, 1), padding='same', kernel_regularizer=l2(0.0002), activation='relu', name='block4_conv3')(x)
    x = MaxPooling2D(pool_size=(2, 2), strides=(2, 2), name="block4_pool", padding='valid')(x)

    # Block 5
    x = Conv2D(filters=512, kernel_size=3, strides=(1, 1), padding='same', kernel_regularizer=l2(0.0002), activation='relu', name='block5_conv1')(x)
    x = Conv2D(filters=512, kernel_size=3, strides=(1, 1), padding='same', kernel_regularizer=l2(0.0002), activation='relu', name='block5_conv2')(x)
    x = Conv2D(filters=512, kernel_size=3, strides=(1, 1), padding='same', kernel_regularizer=l2(0.0002), activation='relu', name='block5_conv3')(x)
    x = MaxPooling2D(pool_size=(2, 2), strides=(2, 2), name="block5_pool", padding='valid')(x)

    # Classification block
    x = Flatten(name='flatten')(x)
    x = Dense(4096, activation='relu', name='fc1')(x)
    x = Dropout(0.5, name='drop_fc1')(x)

    x = Dense(4096, activation='relu', name='fc2')(x)
    x = Dropout(0.5, name='drop_fc2')(x)
    x = Dense(365, activation='softmax', name="predictions")(x)

    # Create model.
    model = Model(img_input, x, name='vgg16-places365')

    # load weights
    weights_path = get_file('vgg16-places365_weights_tf_dim_ordering_tf_kernels.h5', WEIGHTS_PATH, cache_subdir='models')
    model.load_weights(weights_path)

    if K.backend() == 'theano':
        layer_utils.convert_all_kernels_in_model(model)

    if K.image_data_format() == 'channels_first':
        maxpool = model.get_layer(name='block5_pool')
        shape = maxpool.output_shape[1:]
        dense = model.get_layer(name='fc1')
        layer_utils.convert_dense_weights_data_format(dense, shape, 'channels_first')

        if K.backend() == 'tensorflow':
            warnings.warn(
                'You are using the TensorFlow backend, yet you are using the Theano image data format convention (`image_data_format="channels_first"`). '
                'For best performance, set `image_data_format="channels_last"` in your Keras config at ~/.keras/keras.json.'
            )

    return model


def predict(image_file, n_top=5):
    image = Image.open(image_file)
    image = np.array(image, dtype=np.uint8)
    image = resize(image, (224, 224))
    image = np.expand_dims(image, 0)

    model = VGG16_Places365()
    preds = model.predict(image)[0]
    top_preds = np.argsort(preds)[::-1][0:n_top]
    return [CLASS_LABELS[v] for v in top_preds]


